-------------------------------------------------------------------------------
--
-- VGA video generator
--
-- Uses the vga_timing module to generate hsync etc.
-- Massages the RAM address and requests cycles from the memory controller
-- to generate video using one byte per pixel
--
-- Cristian Soviani, Dennis Lim, and Stephen A. Edwards
--
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity opb_xsb300_vga is
	generic (
		C_OPB_AWIDTH		: integer := 32;
		C_OPB_DWIDTH		: integer := 32;
		C_BASEADDR			: std_logic_vector := X"FEFF_0200";
		C_HIGHADDR			: std_logic_vector := X"FEFF_03FF");

	port (
		OPB_Clk : in std_logic;
		OPB_Rst : in std_logic;
		OPB_ABus : in std_logic_vector (31 downto 0);
		OPB_BE : in std_logic_vector (3 downto 0);
		OPB_DBus : in std_logic_vector (31 downto 0);
		OPB_RNW : in std_logic;
		OPB_select : in std_logic;
		OPB_seqAddr : in std_logic;
		UIO_retry : out std_logic := '0';
		UIO_DBus : out std_logic_vector (31 downto 0) := "00000000000000000000000000000000";
		UIO_errAck : out std_logic;
		UIO_toutSup : out std_logic;
		UIO_xferAck : out std_logic;
		pixel_clock : in std_logic;
		VIDOUT_CLK : out std_logic;
		VIDOUT_RCR : out std_logic_vector(9 downto 0);
		VIDOUT_GY : out std_logic_vector(9 downto 0);
		VIDOUT_BCB : out std_logic_vector(9 downto 0);
		VIDOUT_BLANK_N : out std_logic;
		VIDOUT_HSYNC_N : out std_logic;
		VIDOUT_VSYNC_N : out std_logic);
end opb_xsb300_vga;

architecture Behavioral of opb_xsb300_vga is
	-- Fast low-voltage TTL-level I/O pad with 12 mA drive
	
  component OBUF_F_12
		port (
			O : out STD_ULOGIC;
			I : in STD_ULOGIC);
	end component;

	-- Basic edge-sensitive flip-flop
	
	component FD
		port (
			C : in std_logic;
			D : in std_logic;
			Q : out std_logic);
	end component;

	-- Force instances of FD into pads for speed
	
	attribute iob : string;
	attribute iob of FD : component is "true";

	component vga_timing
		port (
			h_sync_delay				 : out std_logic;
			v_sync_delay				 : out std_logic;
			blank								: out std_logic;
			vga_ram_read_addresser : out std_logic_vector (7 downto 0);
			pixel_clock					: in std_logic;
			reset								: in std_logic;
			font_line						: out std_logic_vector(2 downto 0);
                        line_loc : out std_logic_vector(9 downto 0);
			need_reload					: out std_logic
			);
		
	end component;
	

	component fontrom128_64
		port (
		Clk : in std_logic;
		en : in std_logic;
		addr : in std_logic_vector( 15 downto 0 );
		hline : in std_logic_vector( 2 downto 0 );
		data : out std_logic_vector( 15 downto 0 )
	);
	end component;
	
	signal r : std_logic_vector (9 downto 0);
	signal g : std_logic_vector (9 downto 0);
	signal b : std_logic_vector (9 downto 0);
	signal blank : std_logic;
	signal hsync : std_logic;
	signal vsync : std_logic;
	signal vga_ram_read_address : std_logic_vector(7 downto 0);
	signal video_data : std_logic_vector(15 downto 0);
	signal vreq : std_logic;
	signal vreq_1 : std_logic;
	signal load_video_word : std_logic;
	signal vga_shreg : std_logic_vector(15 downto 0);
	signal cur_line : std_logic_vector(2 downto 0);
	signal reload : std_logic;
	signal font_map : std_logic_vector(15 downto 0);
	signal cur_pix : std_logic_vector(2 downto 0);
	signal counter : std_logic_vector(3 downto 0) := "0000";
	signal cs : std_logic;
	signal q0 : std_logic;
	signal d0 : std_logic;
	signal xfer : std_logic;
	signal wdata : std_logic_vector( 7 downto 0 );
	signal addr : std_logic_vector( 8 downto 0 );
        signal yloc:std_logic_vector (9 downto 0);
        signal mask : std_logic_vector(2 downto 0);
        signal mask2 : std_logic_vector(2 downto 0);
--  signal beg_time_reg : std_logic_vector( 8 downto 0 );
--  signal end_time_reg : std_logic_vector( 8 downto 0 );
  signal write_enable : std_logic;
--  signal color_mask : std_logic_vector( 2 downto 0 ) := "111";
begin
	st : vga_timing port map (
		pixel_clock => pixel_clock,
		reset => OPB_Rst,
		h_sync_delay => hsync,
		v_sync_delay => vsync,
		blank => blank,
		vga_ram_read_addresser => vga_ram_read_address,
		font_line => cur_line,
		need_reload => reload,
                line_loc => yloc
	);

	buff : RAMB4_S8_S16 port map (
		-- write
		CLKA => OPB_Clk,
		RSTA => OPB_Rst,
		DIA => wdata,
		ENA => '1',
		WEA => write_enable,
		ADDRA => addr,
		DOA => open,
		
		-- read
		CLKB => pixel_clock,
		RSTB => OPB_Rst,
		DIB => X"0000",
		ENB => '1',
		WEB => '0',
		ADDRB => vga_ram_read_address( 7 downto 0 ),
		DOB => video_data
	);

	ft : fontrom128_64 port map (
		Clk=>pixel_clock,
		en=>'1',
		addr=>video_data,
		hline=>cur_line,
		data=>font_map
	);

	UIO_DBus <= X"0000_0000";
	cs <= OPB_select when ( OPB_ABus(31 downto 8) = X"FEFF02" or OPB_ABus( 31 downto 8 ) = X"FEFF03" ) else '0';
        
	process( OPB_Clk, OPB_Rst )
	begin
		if OPB_Rst = '1' then
			d0 <= '0';
		elsif OPB_Clk'event and OPB_Clk = '1' then
			d0 <= ( not q0 ) and cs;
			q0 <= d0;
		end if;
	end process;

	xfer <= q0;
	write_enable <= q0;

	process( OPB_Clk )
	begin
		if OPB_Clk'event and OPB_Clk = '1' then
		  --wdata <= OPB_DBus( 31 downto 24 );
                  wdata <= OPB_DBus( 7 downto 0 );      
		end if;
	end process;

	process( OPB_Clk )
	begin
		if OPB_Clk'event and OPB_Clk = '1' then
			--addr <= OPB_ABus( 9 downto 1 );
              		addr <= OPB_ABus( 8 downto 0 );
		end if;
	end process;

--        process( OPB_Clk )
--        begin
--         if OPB_Clk'event and OPB_Clk = '1' then 
--            if vga_ram_read_address( 7 downto 0 ) = X"FE" then
--              beg_time_reg <= video_data( 0 ) & video_data( 15 downto 8 );
--            elsif vga_ram_read_address( 7 downto 0 ) = X"FF" then
--              end_time_reg <= video_data( 0 ) & video_data( 15 downto 8 );
--            end if;
--          end if;
--        end process;
        
	-- Video request is true when the RAM address is even
	
	-- FIXME: This should be disabled during blanking to reduce memory traffic

	--vreq <= reload ;--not vga_ram_read_address(0);

--	vreq <= not vga_ram_read_address(0);
	-- Generate load_video_word by delaying vreq two cycles
	
--	process (pixel_clock)
--	begin
	--	if pixel_clock'event and pixel_clock='1' then
		--	 if counter="1111" then
			--	 load_video_word <= '1';

                          --       counter <= "0000";
			--else
		--		 load_video_word <='0';
			--	 counter <= counter+1;
	--		end if;
		
--		end if;
--	end process;

	-- Generate video_req (to the RAM controller) by delaying vreq by
	-- a cycle synchronized with the pixel clock
	
--	process (clk)
--	begin
--		if clk'event and clk='1' then
--			video_req <= pixel_clock ;--and vreq;
--		end if;
--	end process;

	-- The video address is the upper 19 bits from the VGA timing generator
	-- because we are using two pixels per word and the RAM address counts words

	--video_addr <=	vga_ram_read_address(19 downto 0);

	-- The video shift register: either load it from RAM or shift it up a byte
	
	process (pixel_clock)
	begin
		if pixel_clock'event and pixel_clock='1' then
			if reload = '1' then
			--if reload = '1' then
				vga_shreg <= font_map(7 downto 0) & font_map(15 downto 8); --video_data;
				-- Shift the low byte of read video data into the high byte
			else
                          vga_shreg <= vga_shreg(14 downto 0) & "0";
			end if;
                        if yloc>255 and yloc<304 then
                          mask <= "111";
                        else
                          mask <= "000";                          
                        end if;
                        if yloc>255 and yloc <264 then
                          mask2 <="000";
                        else
                          mask2 <= "111";
                        end if;
		end if;
	end process;

--        process( pixel_clock )
--        begin
--          if pixel_clock'event and pixel_clock = '1' then
--            if vga_ram_read_address( 8 downto 0 ) = beg_time_reg then
--              color_mask <= "000";
--            elsif vga_ram_read_address( 8 downto 0 ) = end_time_reg then
--              color_mask <= "111";
--            end if;
--          end if;
--        end process;
        
	-- Copy the upper byte of the video word to the color signals
	-- Note that we use three bits for red and green and two for blue.

	r(9 downto 7) <= vga_shreg(15)&vga_shreg(15)&vga_shreg(15) and mask; -- and color_mask;
	r(6 downto 0) <= "0000000";
	g(9 downto 7) <= (vga_shreg(15)&vga_shreg(15)&vga_shreg(15) and mask) and mask2;
	g(6 downto 0) <= "0000000";
	b(9 downto 8) <= vga_shreg(15)&vga_shreg(15) and mask(1 downto 0); -- and color_mask(
	b(7 downto 0) <= "00000000";

	-- Video clock I/O pad to the DAC
	
	vidclk : OBUF_F_12 port map (
		O => VIDOUT_clk,
		I => pixel_clock);

	-- Control signals: hsync, vsync, and blank

	hsync_ff : FD port map (
		C => pixel_clock,
		D => not hsync,
		Q => VIDOUT_HSYNC_N );

	vsync_ff : FD port map (
		C => pixel_clock,
		D => not vsync,
		Q => VIDOUT_VSYNC_N );

	blank_ff :	FD port map (
		C => pixel_clock,
		D => not blank,
		Q => VIDOUT_BLANK_N );

	-- Three digital color signals
	
	rgb_ff : for i in 0 to 9 generate
		
		r_ff : FD port map (
			C => pixel_clock,
			D => r(i),
			Q => VIDOUT_RCR(i) );
		
		g_ff : FD port map (
			C => pixel_clock,
			D => g(i),
			Q => VIDOUT_GY(i) );
		
		b_ff : FD port map (
			C => pixel_clock,
			D => b(i),
			Q => VIDOUT_BCB(i) );
		
	end generate;

	UIO_errAck <= '0';
	UIO_retry <= '0';
	UIO_toutSup <= '0';
	UIO_xferAck <= xfer;
	
end Behavioral;
